#include "CloudData.h"
#include "Config.h"
#include "Matrices.h"
#include <math.h>
#include <assert.h>

CloudData::CloudData() : noofbuffers(0)
{
  width=height=depth=1;
  starts[0]=starts[1]=starts[2]=0;
  
  int noofregions=width*height*depth;
  regions=(REGION**) malloc(4*noofregions);
  memset(regions, 0, 4*noofregions);

  templistsize=NULL;
  templist=NULL;

  SetNoOfBuffers(1);
}

CloudData::~CloudData()
{
  int size=width*height*depth;
  for (int i=0; i<size; i++)
  {
    if (regions[i])
    {
      free(regions[i]->points);
      free(regions[i]);
    }
  }
  free(regions);

  for (int i=0; i<noofbuffers; i++)
    free(templist[i]);
  free(templist);
  free(templistsize);
}

void CloudData::SetNoOfBuffers(const int n)
{
  if (n!=noofbuffers)
  {
    noofbuffers=n;
    templistsize=(int*) realloc(templistsize, sizeof(int)*n);
    templist=(POINTDATA***) realloc(templist, 4*n);
    for (int i=0; i<n; i++)
    {
      templistsize[i]=0;
      templist[i]=NULL;
    }
  }
}

void CloudData::ClearPointData()
{
  int noofregions=width*height*depth;
  for (int i=0; i<noofregions; i++)
  {
    if (regions[i])
      regions[i]->noofpts=0;
  }
}

void CloudData::GetIndicies(int* indicies, const float x, const float y, const float z)
{
  float t=x*REGION_SCALE;
  indicies[0]=(int) (t<0?t-1:t)-starts[0];

  t=y*REGION_SCALE;
  indicies[1]=(int) (t<0?t-1:t)-starts[1];

  t=z*REGION_SCALE;
  indicies[2]=(int) (t<0?t-1:t)-starts[2];
}

void CloudData::AddPointData(POINTDATA* point)
{
  int txyz[3];
  GetIndicies(txyz, point->xyz[0], point->xyz[1], point->xyz[2]);

  if (txyz[0]<0)
  {
    ResizeRegionsX(txyz[0]);
    txyz[0]=0;
  }
  else if (txyz[0]>=width)
    ResizeRegionsX(txyz[0]-width+1);

  if (txyz[1]<0)
  {
    ResizeRegionsY(txyz[1]);
    txyz[1]=0;
  }
  else if (txyz[1]>=height)
    ResizeRegionsY(txyz[1]-height+1);

  if (txyz[2]<0)
  {
    ResizeRegionsZ(txyz[2]);
    txyz[2]=0;
  }
  else if (txyz[2]>=depth)
    ResizeRegionsZ(txyz[2]-depth+1);

//  int t2xyz[3];
//  GetIndicies(t2xyz, point->xyz[0], point->xyz[1], point->xyz[2]);
//  assert(txyz[0]==t2xyz[0] && txyz[1]==t2xyz[1] && txyz[2]==t2xyz[2]);

  REGION* region=NULL;
  int index=txyz[0]+width*(txyz[1]+txyz[2]*height);
  if (regions[index])
    region=regions[index];
  else
  {
    region=regions[index]=(REGION*) malloc(sizeof(REGION));
    memset(region, 0, sizeof(REGION));
  }

  if (region->noofpts==region->maxpts)
  {
    region->maxpts+=100;
    region->points=(POINTDATA**) realloc(region->points, 4*region->maxpts);
  }
  region->points[region->noofpts++]=point;
}

void CloudData::ResizeRegionsX(int amount)
{
  int add=(amount<0?-amount:amount);
  int nwidth=width+add;
    
  int noofregions=nwidth*height*depth;
  REGION** tmp=(REGION**) malloc(4*noofregions);
  memset(tmp, 0, 4*noofregions);
  for (int k=0; k<depth; k++)
  {
    int tds=k*width*height;
    int tdd=k*nwidth*height;
    if (amount<0)
      tdd+=add;
    for (int j=0; j<height; j++)
    {
      int ths=tds+j*width;
      int thd=tdd+j*nwidth;
      for (int i=0; i<width; i++)
        tmp[thd+i]=regions[ths+i];
    }
  }

  free(regions);
  regions=tmp;
  width=nwidth;
  if (amount<0)
    starts[0]+=amount;
}

void CloudData::ResizeRegionsY(int amount)
{
  int add=(amount<0?-amount:amount);
  int nheight=height+add;
    
  int noofregions=width*nheight*depth;
  REGION** tmp=(REGION**) malloc(4*noofregions);
  memset(tmp, 0, 4*noofregions);
  for (int k=0; k<depth; k++)
  {
    int tds=k*width*height;
    int tdd=k*width*nheight;
    if (amount<0)
      tdd+=add*width;
    for (int j=0; j<height; j++)
    {
      int ths=tds+j*width;
      int thd=tdd+j*width;
      for (int i=0; i<width; i++)
        tmp[thd+i]=regions[ths+i];
    }
  }

  free(regions);
  regions=tmp;
  height=nheight;
  if (amount<0)
    starts[1]+=amount;
}

void CloudData::ResizeRegionsZ(int amount)
{
  int add=(amount<0?-amount:amount);
  int ndepth=depth+add;
    
  int noofregions=width*height*ndepth;
  REGION** tmp=(REGION**) malloc(4*noofregions);
  memset(tmp, 0, 4*noofregions);
  for (int k=0; k<depth; k++)
  {
    int tds=k*width*height;
    int tdd=k*width*height;
    if (amount<0)
      tdd+=add*width*height;
    for (int j=0; j<height; j++)
    {
      int ths=tds+j*width;
      int thd=tdd+j*width;
      for (int i=0; i<width; i++)
        tmp[thd+i]=regions[ths+i];
    }
  }

  free(regions);
  regions=tmp;
  depth=ndepth;
  if (amount<0)
    starts[2]+=amount;
}

bool CloudData::CheckConsistency()
{
  int index=0;
  for (int k=0; k<depth; k++)
  {
    for (int j=0; j<height; j++)
    {
      for (int i=0; i<width; i++)
      {
        REGION* region=regions[index++];
        if (region)
        {
          float mins[3], maxs[3];
          mins[0]=(starts[0]+i)/REGION_SCALE;
          mins[1]=(starts[1]+j)/REGION_SCALE;
          mins[2]=(starts[2]+k)/REGION_SCALE;

          maxs[0]=mins[0]+1/REGION_SCALE;
          maxs[1]=mins[1]+1/REGION_SCALE;
          maxs[2]=mins[2]+1/REGION_SCALE;

          for (int l=0; l<region->noofpts; l++)
          {
            for (int m=0; m<3; m++)
            {
              if (!(region->points[l]->xyz[m]>=mins[m] && region->points[l]->xyz[m]<=maxs[m]))
              {
                assert(false);
                return false;
              }
            }
          }
        }
      }
    }
  }
  return true;
}

POINTDATA** CloudData::GetPoints(float* min, float* max, int& noofpts, const int bufferno, const bool reducewindow)
{
  assert(min[0]<max[0] && min[1]<max[1] && min[2]<max[2]);

  int minxyz[3], maxxyz[3];
  GetIndicies(minxyz, min[0], min[1], min[2]);
  GetIndicies(maxxyz, max[0], max[1], max[2]);

  if (reducewindow)
  {
    for (int i=0; i<3; i++)
    {
      ++minxyz[i];
      --maxxyz[i];
    }
  }
  ClampIndicies(minxyz);
  ClampIndicies(maxxyz);

  if (reducewindow)
  {
    min[0]=(starts[0]+minxyz[0])/REGION_SCALE;
    min[1]=(starts[1]+minxyz[1])/REGION_SCALE;
    min[2]=(starts[2]+minxyz[2])/REGION_SCALE;

    max[0]=(1+starts[0]+maxxyz[0])/REGION_SCALE;
    max[1]=(1+starts[1]+maxxyz[1])/REGION_SCALE;
    max[2]=(1+starts[2]+maxxyz[2])/REGION_SCALE;
  }

  return GetPointsAux(minxyz, maxxyz, noofpts, bufferno);
}

POINTDATA** CloudData::GetPoints(const float x, const float y, const float z, const float tolerance, int& noofpts, const int bufferno)
{
  int minxyz[3], maxxyz[3];
  GetIndicies(minxyz, x-tolerance, y-tolerance, z-tolerance);
  GetIndicies(maxxyz, x+tolerance, y+tolerance, z+tolerance);

  ClampIndicies(minxyz);
  ClampIndicies(maxxyz);

  return GetPointsAux(minxyz, maxxyz, noofpts, bufferno);
}

POINTDATA** CloudData::GetPointsAux(int* minxyz, int* maxxyz, int& noofpts, const int bufferno)
{
  assert(bufferno<noofbuffers);

  noofpts=0;

  int templistsize=this->templistsize[bufferno];
  POINTDATA** templist=this->templist[bufferno];

  for (int k=minxyz[2]; k<=maxxyz[2]; k++)
  {
    for (int j=minxyz[1]; j<=maxxyz[1]; j++)
    {
      for (int i=minxyz[0]; i<=maxxyz[0]; i++)
      {
        assert(i>=0 && i<width && j>=0 && j<height && k>=0 && k<depth);

        int index=i+j*width+k*width*height;

        REGION* region=regions[index];
        if (region)
        {
          if (templistsize<noofpts+region->noofpts)
          {
            this->templistsize[bufferno]+=region->noofpts;
            this->templist[bufferno]=(POINTDATA**) realloc(this->templist[bufferno], 4*this->templistsize[bufferno]);

            templistsize=this->templistsize[bufferno];
            templist=this->templist[bufferno];
          }
          for (int l=0; l<region->noofpts; l++)
            templist[noofpts++]=region->points[l];
        }
      }
    }
  }

  return templist;
}

POINTDATA** CloudData::GetPoints(const int noofcentres, const float* centres, const float radius, int& noofpts, const int bufferno)
{
  noofpts=0;

  int noofregions=width*height*depth;
  bool* used=(bool*) malloc(sizeof(bool)*noofregions);
  memset(used, 0, sizeof(bool)*noofregions);

  int templistsize=this->templistsize[bufferno];
  POINTDATA** templist=this->templist[bufferno];

  const float rsqr=radius*radius;

  for (int c=0; c<noofcentres; c++)
  {
    int minxyz[3], maxxyz[3];
    GetIndicies(minxyz, centres[0]-radius, centres[1]-radius, centres[2]-radius);
    GetIndicies(maxxyz, centres[0]+radius, centres[1]+radius, centres[2]+radius);

    ClampIndicies(minxyz);
    ClampIndicies(maxxyz);

    for (int k=minxyz[2]; k<=maxxyz[2]; k++)
    {
      for (int j=minxyz[1]; j<=maxxyz[1]; j++)
      {
        for (int i=minxyz[0]; i<=maxxyz[0]; i++)
        {
          assert(i>=0 && i<width && j>=0 && j<height && k>=0 && k<depth);

          int index=i+j*width+k*width*height;
          REGION* region=regions[index];

          if (region && !used[index])
          {
            float mins[3], maxs[3], dxyz[3];
            mins[0]=(starts[0]+i)/REGION_SCALE;
            mins[1]=(starts[1]+j)/REGION_SCALE;
            mins[2]=(starts[2]+k)/REGION_SCALE;

            maxs[0]=mins[0]+1/REGION_SCALE;
            maxs[1]=mins[1]+1/REGION_SCALE;
            maxs[2]=mins[2]+1/REGION_SCALE;

            for (int l=0; l<3; l++)
            {
              float diff1=mins[l]-centres[l];
              float diff2=maxs[l]-centres[l];
              dxyz[l]=(fabs(diff1)<fabs(diff2)?diff1:diff2);
            }

            float distsqr=dxyz[0]*dxyz[0]+dxyz[1]*dxyz[1]+dxyz[2]*dxyz[2];
            if (distsqr<rsqr)
            {
              if (templistsize<noofpts+region->noofpts)
              {
                this->templistsize[bufferno]+=region->noofpts;
                this->templist[bufferno]=(POINTDATA**) realloc(this->templist[bufferno], 4*this->templistsize[bufferno]);

                templistsize=this->templistsize[bufferno];
                templist=this->templist[bufferno];
              }
              for (int l=0; l<region->noofpts; l++)
                templist[noofpts++]=region->points[l];

              used[index]=true;
            }
          }
        }
      }
    }
    centres+=4;
  }

  free(used);

  return templist;
}

POINTDATA** CloudData::GetPoints(float* min, float* max, const double* inverse, const float* f_near, const float* f_far, int& noofpts, const int bufferno, const bool reducewindow)
{
  noofpts=0;

  int minxyz[3], maxxyz[3];
  GetIndicies(minxyz, min[0], min[1], min[2]);
  GetIndicies(maxxyz, max[0], max[1], max[2]);
/*
  if (reducewindow)
  {
    for (int i=0; i<3; i++)
    {
      ++minxyz[i];
      --maxxyz[i];
    }
  }
  */
  ClampIndicies(minxyz);
  ClampIndicies(maxxyz);
/*
  if (reducewindow)
  {
    min[0]=(starts[0]+minxyz[0])/REGION_SCALE;
    min[1]=(starts[1]+minxyz[1])/REGION_SCALE;
    min[2]=(starts[2]+minxyz[2])/REGION_SCALE;

    max[0]=(1+starts[0]+maxxyz[0])/REGION_SCALE;
    max[1]=(1+starts[1]+maxxyz[1])/REGION_SCALE;
    max[2]=(1+starts[2]+maxxyz[2])/REGION_SCALE;
  }
  */

  int templistsize=this->templistsize[bufferno];
  POINTDATA** templist=this->templist[bufferno];

  const float rangex=f_far[0]-f_near[0];
  const float rangey=f_far[1]-f_near[1];
  const float rangez=f_near[2]-f_far[2];

  for (int k=minxyz[2]; k<=maxxyz[2]; k++)
  {
    for (int j=minxyz[1]; j<=maxxyz[1]; j++)
    {
      for (int i=minxyz[0]; i<=maxxyz[0]; i++)
      {
        assert(i>=0 && i<width && j>=0 && j<height && k>=0 && k<depth);

        float tmpvec[4], mins[4], maxs[4];
        tmpvec[0]=(starts[0]+i)/REGION_SCALE;
        tmpvec[1]=(starts[1]+j)/REGION_SCALE;
        tmpvec[2]=(starts[2]+k)/REGION_SCALE;
        tmpvec[3]=1.0f;
        MatMult(mins, inverse, 4, 4, tmpvec, 4, 1);

        tmpvec[0]+=1/REGION_SCALE;
        tmpvec[1]+=1/REGION_SCALE;
        tmpvec[2]+=1/REGION_SCALE;
        MatMult(maxs, inverse, 4, 4, tmpvec, 4, 1);

        REGION* region=NULL;

        bool tmin=(mins[2]>f_far[2] && mins[2]<f_near[2]);
        bool tmax=(maxs[2]>f_far[2] && maxs[2]<f_near[2]);

        int index=-1;
        if (tmin)
        {
          float position=f_near[2]-mins[2];
          float scale=position/rangez;
          assert(scale>=0.0f && scale<=1.0f);

          float tx=f_near[0]+rangex*scale;
          float ty=f_near[1]+rangey*scale;

          if (mins[0]>-tx && mins[0]<tx && mins[1]>-ty && mins[1]<ty)
          {
            index=i+j*width+k*width*height;
            region=regions[index];
          }
        }

        if (tmax && index==-1)
        {
          float position=f_near[2]-maxs[2];
          float scale=position/rangez;
          assert(scale>=0.0f && scale<=1.0f);

          float tx=f_near[0]+rangex*scale;
          float ty=f_near[1]+rangey*scale;

          if (maxs[0]>-tx && maxs[0]<tx && maxs[1]>-ty && maxs[1]<ty)
          {
            index=i+j*width+k*width*height;
            region=regions[index];
          }
        }

        if (region)
        {
          if (templistsize<noofpts+region->noofpts)
          {
            this->templistsize[bufferno]+=region->noofpts;
            this->templist[bufferno]=(POINTDATA**) realloc(this->templist[bufferno], 4*this->templistsize[bufferno]);

            templistsize=this->templistsize[bufferno];
            templist=this->templist[bufferno];
          }
          for (int l=0; l<region->noofpts; l++)
            templist[noofpts++]=region->points[l];
        }
      }
    }
  }

  return templist;
}

void CloudData::CheckPrune(POINTDATA* wlddata, int& noofwldpts, POINTDATA* wldpt, const int iteration)
{
  DWORD crop=iteration-DROP_RATE;//max-30;//min+diff/10;

  if (iteration<DROP_RATE || wldpt->lastseen>=crop)
    return;

  int indicies[3];
  GetIndicies(indicies, wldpt->xyz[0], wldpt->xyz[1], wldpt->xyz[2]);

  int index=indicies[0]+width*(indicies[1]+indicies[2]*height);
  assert(index<width*height*depth);
  REGION* region=regions[index];
  assert(region);

  bool found=false;
  for (int j=0; j<region->noofpts && !found; j++)
  {
    if (region->points[j]==wldpt)
    {
      int index=region->points[j]-wlddata;

      // find wlddata[noofwldpts-1] in region info and update with wlddata[index]
      int txyz[3];
      POINTDATA* search=wlddata+noofwldpts-1;
      GetIndicies(txyz, search->xyz[0], search->xyz[1], search->xyz[2]);
      REGION* subregion=regions[txyz[0]+width*(txyz[1]+txyz[2]*height)];
      for (int k=0; k<subregion->noofpts && !found; k++)
      {
        if (subregion->points[k]==search)
        {
          subregion->points[k]=wlddata+index;
          found=true;
        }
      }
      assert(found);

      wlddata[index]=*search;//wlddata[noofwldpts-1];
      --noofwldpts;

      region->points[j]=region->points[region->noofpts-1];
      --region->noofpts;
      --j;
    }
  }

  assert(found);
}

void CloudData::Prune(POINTDATA* wlddata, int& noofwldpts, const float* minxyz, const float* maxxyz, const int iteration)
{
  if (iteration<DROP_RATE)
    return;
//  if (true)
//    return;

  assert(minxyz[0]<maxxyz[0] && minxyz[1]<maxxyz[1] && minxyz[2]<maxxyz[2]);

  int minxyzids[3], maxxyzids[3];
  GetIndicies(minxyzids, minxyz[0], minxyz[1], minxyz[2]);
  GetIndicies(maxxyzids, maxxyz[0], maxxyz[1], maxxyz[2]);

  // don't prune the edges some of the points might be out of shot
  for (int i=0; i<3; i++)
  {
    ++minxyzids[i];
    --maxxyzids[i];
  }
  
  ClampIndicies(minxyzids);
  ClampIndicies(maxxyzids);

  DWORD crop=iteration-DROP_RATE;//max-30;//min+diff/10;

  for (int z=minxyzids[2]; z<=maxxyzids[2]; z++)
  {
    for (int y=minxyzids[1]; y<=maxxyzids[1]; y++)
    {
      for (int x=minxyzids[0]; x<=maxxyzids[0]; x++)
      {
        REGION* region=regions[x+width*(y+z*height)];
        if (region && region->noofpts>1)
        {
/*          DWORD min=region->points[0]->lastseen;
          DWORD max=min;
          for (int j=1; j<region->noofpts; j++)
          {
            if (region->points[j]->lastseen<min)
              min=region->points[j]->lastseen;
            else if (region->points[j]->lastseen>max)
              max=region->points[j]->lastseen;
          }*/

          //DWORD diff=max-min;
          //if (diff>30)
          {
//            DWORD crop=DROP_RATE-30;//max-30;//min+diff/10;
            for (int j=0; j<region->noofpts; j++)
            {
              if (region->points[j]->lastseen<crop)
              {
                int index=region->points[j]-wlddata;

                // find wlddata[noofwldpts-1] in region info and update with wlddata[index]
                int txyz[3];
                POINTDATA* search=wlddata+noofwldpts-1;
                GetIndicies(txyz, search->xyz[0], search->xyz[1], search->xyz[2]);
                REGION* subregion=regions[txyz[0]+width*(txyz[1]+txyz[2]*height)];
                bool found=false;
                for (int k=0; k<subregion->noofpts && !found; k++)
                {
                  if (subregion->points[k]==search)
                  {
                    subregion->points[k]=wlddata+index;
                    found=true;
                  }
                }
                assert(found);

                wlddata[index]=*search;//wlddata[noofwldpts-1];
                --noofwldpts;

                region->points[j]=region->points[region->noofpts-1];
                --region->noofpts;
                --j;
              }
            }
          }
        }
      }
    }
  }
}