#pragma once
#include <Windows.h>
#include <math.h>
#include <NuiApi.h>
#include <stdio.h>
#include "Button.h"
#include "CloudData.h"
#include "Config.h"
#include "GLRenderer.h"
#include "Matrices.h"
#include "PairMatch.h"
#include "PDUpdate.h"
#include "Recorder.h"

struct PREVIEWINFO
{
  int x, y, width, height;
  int srcwidth, srcheight;
  DWORD* data;
  HBITMAP hbmpprev;

  int fps;
  DWORD counter;
  wchar_t lastfps[20];

  POINT calibration[2];
};

struct PAIR
{
//  int id1, id2;
  POINTDATA* pt1, *pt2;
//  float error;
};

struct JACOBIANTHREADINFO
{
  int offset, noofpairs, totalpairs;
  const PAIR* pairs;
  const double* coss, *sins;
  double* jacobian;
//  double t6, t7, t8, t9, t10, t11, t9_2, t6_2;

  bool* alive;
  HANDLE alerts[2];
};

struct ICPERRORTHREADINFO
{
  int noofpairs;
  const PAIR* pairs;
  
  double* r;
  const double* transform;

  double error;

  bool* alive;
  HANDLE alerts[2];
};

class MainWnd
{
public:
  MainWnd(HINSTANCE hInstance);
  ~MainWnd();

protected:
  void Init(HWND hwnd);
  void OnSize(UINT nType, int cx, int cy);
  void OnDraw();
  void OnLButtonDown(UINT nFlags, int x, int y);
  void OnMouseMove(UINT nFlags, int x, int y);
  void OnLButtonUp(UINT nFlags, int x, int y);
  void OnMouseWheel(WPARAM wParam, int x, int y);
  void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
  void OnGLRedraw(LPARAM lParam);
  void OnTimer(UINT nIDEvent);

  void WaitForData();

  friend LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
  friend DWORD WINAPI ListenThread(LPVOID lpParameter);
  friend DWORD WINAPI PlaybackThread(LPVOID lpParameter);

private:
  bool kinectalive, tracking, grabdata;
  HINSTANCE hinstance;
  HWND hwnd;
  GLRenderer renderer[2];
  HWND infos[2];
  Button setangle[2], track, swap, reset, record, playback, grab;

  int trackedframes, lastgrab, lastpruned;

  bool exportdata[2];

  bool recording, playing;
  Recorder* recorder;
/*
  bool* depthseen;
  float* depthmins, *depthmaxs, *depthavgs;
  int* depthcounts;
  int depthframes;
  */
  HDC hdctmp;
  PREVIEWINFO colprev, depthprev;

  CloudData clouddata;
  int noofwldpts, maxwldpts;
  POINTDATA* pointdata, *worlddata;
  CAMERATRACK cameratrack;
  float calibration[9];  // from depth space into colour space

//  float framemin[3], framemax[3];

  INuiSensor* sensor;
  HANDLE colstream, depthstream;
  HANDLE frameevents[2]; // 0 is colour and 1 is depth
  WORD depthrange[3];
  long sensorangle;

  // Tracking Memory
  PairMatch pairmatch;
  PDUpdate pdupdate;

  PAIR* pairs;
  double* r, /**update, */*jacobian, *tmpsqr, *lowermat, *b, *y;
  float pairratio[2]; // 0=the number of pairs/used 1=the number of pairs within the min distance threshold/pairs

  bool threadalive;
  int noofthreads;
  HANDLE* jacobthreads, *icperrorthreads, *alerts;
  JACOBIANTHREADINFO* jacobthreadinfo;
  ICPERRORTHREADINFO* icpthreadinfo;

  void SetNoOfThreads(int n);
  void KillThreads();

  HRESULT InitKinect();
  void ProcessColour();
  void ProcessDepth();
  void UpdateFPS(PREVIEWINFO& info);
  void DrawAngleControls(HDC hdc, bool clear);
  void SetCalibrationPoint(PREVIEWINFO& preview, int x, int y, bool second);
  void UpdateCalibration();
  void DrawCalibrationPoints(HDC hdc, PREVIEWINFO& preview);
  void SaveSettings();
  void LoadSettings();
  void SwapViews();
  void ResetGlobal();
  void TrackData();
  void Record();
  void Playback();
  void GrabData();

  void CalcNormals(POINTDATA* frame, int& used);

  float CalcDataDepth(const WORD depth);
  void TrackData(POINTDATA* frame, int used);
  void TrackDataAux(POINTDATA* frame, const int used);
  double TrackDataOuter(POINTDATA* frame, const int used);
  void TrackDataInner(POINTDATA* frame, const int used, const int noofpairs);

  void CalcCosSins(double* coss, double* sins, const double* rxyz);
  void CalcFrustum(float* f_near, float* f_far);
  void CalcFrameMinMax(const double* transform, float* min, float* max);

  void SavePairs(const POINTDATA* frame, const int framesize, POINTDATA** wldpts, const int noofwldpts, const double* transform, const bool toworld);

  void InitICPErrorThreadMemory(const double* transform, const int noofpairs, const PAIR* pairs);
  void InitJacobianThreadMemory(const double* coss, const double* sins, const int noofpairs, const PAIR* pairs);
  void CalcJacobian();
  double CalcInnerICPError();

  void ClearInfoText(const int index);
  void AppendInfoText(const wchar_t* text, const int index);
  
  void GetTime(wchar_t* buffer, const int bufflen);
  void WriteCameraTrackInfo(wchar_t* buffer, const int bufflen);
  int FormatNumber(wchar_t* buffer, const int bufflen, double value, int dps);
  
  void InitTrackingMemory();
  void FreeTrackingMemory();

  void ScreenGrab();
  void ExportWRLModel(bool frame);
  void CreatePolygonModel(bool frame);

//  void DumpDepthAnalysis();
  
  void TestCode();
  void TestCodeV2();
};

inline float MainWnd::CalcDataDepth(const WORD depth)
{
  return (-depth/1000.0f);   // depth value is given in meters from the camera
//  return Z_OFFSET-depth*WORLD_SCALE*DEPTH_SCALE;
}

inline void MainWnd::CalcCosSins(double* coss, double* sins, const double* rxyz)
{
  for (int i=0; i<3; i++)
  {
    coss[i]=cos(rxyz[i]);
    sins[i]=sin(rxyz[i]);
  }
}